/*
 * ust.h
 *
 * C interface for the universal security transform
 *
 * David A. McGrew
 * Cisco Systems, Inc.
 */


#ifndef UST_H
#define UST_H

#include "crypto_kernel.h"  /* interface to cryptographic kernel      */
#include "rdbx.h"           /* extended seq num and replay database   */
#include "config.h"         /* for HAVE_UIO_H                         */

typedef struct ust_ctx_t *ust_ptr_t;

/*
 * ust_alloc() allocates a ust_ctx_t 
 * (parameter definitions are described below)
 * 
 * auth_type_id_t and cipher_type_id_t are integer identifiers
 * for crypto algorithm types.  statically defined values are
 * defined in crypto_types.h
 */

err_status_t
ust_alloc(ust_ptr_t ctx,             /* ust context                       */
         cipher_type_id_t c_id,      /* keystream generator identifier    */
         int cipher_key_len,         /* number of octets in cipher key    */
         auth_type_id_t a_id,        /* auth algorithm identifier         */
         int auth_key_len,           /* number of octets in the auth key  */
         int auth_tag_len,           /* number of octets in the auth tag  */
         int replay_window_len       /* bits in replay window (0 == none) */
         ); 

/*  
 * ust_init() initializes a ust_ctx_t  
 *
 * the context is initialized with the cipher_key and auth_key
 * provided; the lengths of these keys (in octets) can be found using
 * the functions ust_get_cipher_key_len() and ust_get_auth_key_len().
 * 
 * if auth_key is NULL, then ust_init() will generate the auth_key 
 * using the cipher key (as described in the UST spec)
 */

err_status_t
ust_init(ust_ptr_t ctx, 
         octet_t *cipher_key, 
         octet_t *auth_key
         );


/*
 * the ust_xfm() function 
 *
 * ctx is the ust context, which holds the cipher and hash function
 * keys and parameters, as well as anti-replay information
 * 
 * idx is the packet index, a 48-bit unsigned integer which 
 * should be unique for each invocation of ust_xfm for a
 * given ctx
 * 
 * if auth_start != NULL, then authentication is provided to the
 * auth_len octets of data at auth_start by computing the
 * authentication tag and writing it to *tag; otherwise, the
 * authentication tag is not computed
 *
 * if enc_start != NULL, then encryption is provided to the
 * enc_len octets of data at enc_start by exoring keystream
 * into that data; otherwise, no encryption is done
 *
 * tag points to the authentication tag; after ust_xfm returns,
 * it contains the tag corresonding to the data at auth_start
 * (if auth_start != NULL) -- note that there MUST be at least
 * ust_tag_len(ctx) octets of storage at *tag! 
 */

err_status_t
ust_xfm(ust_ptr_t ctx,       /* pointer to ust context           */
        xtd_seq_num_t idx,   /* index                            */
        octet_t *enc_start,  /* pointer to encryption start      */
        int enc_len,         /* number of octets to encrypt      */
        octet_t *auth_start, /* pointer to authentication start  */
        int auth_len,        /* number of octets to authenticate */
        octet_t *tag         /* authentication tag               */
        );

/*
 * the ust_inv_xfm() function (ust inverse transform)
 *
 * ctx is the ust context, which holds the cipher and hash function
 * keys and parameters, as well as anti-replay information
 * 
 * idx is the packet index, a 64-bit unsigned integer which 
 * should be unique for each invocation of ust_xfm for a
 * given ctx
 * 
 * if auth_start != NULL, then authentication is expected on the
 * auth_len octets of data at auth_start by computing the
 * authentication tag and comparing it to the value at *tag;
 * otherwise, no authentication check is performed
 *
 * if enc_start != NULL, then decryption is done to the enc_len
 * octets of data at enc_start by exoring keystream into that data;
 * otherwise, no decryption is done
 *
 * tag points to the authentication tag;  if auth_start != NULL,
 * then *tag is expected to hold the authentication tag corresponding
 * to the data at *auth_start -- note that there MUST be at least
 * ust_tag_len(ctx) octets of readable data at *tag! 
 */

err_status_t
ust_inv_xfm(ust_ptr_t ctx,       /* ust context                      */
            xtd_seq_num_t idx,   /* index                            */
            octet_t *enc_start,  /* pointer to encryption start      */
            int enc_len,         /* number of octets to encrypt      */
            octet_t *auth_start, /* pointer to authentication start  */
            int auth_len,        /* number of octets to authenticate */
            octet_t *tag         /* authentication tag               */
            );

/*
 * ust_tag_len(ctx) returns the length (in octets) of the
 * authentication tag for the ust context ctx.
 *
 * this function can be used to determine the storage
 * space required to hold a particular tag, if need be
 * 
 * this tag length value is constant for a given ust context, across
 * all invocations of ust_xfm()
 */

unsigned int 
ust_get_tag_len(ust_ptr_t ctx);

/*
 * ust_set_diversifier()
 */

inline err_status_t
ust_set_diversifier(ust_ptr_t ctx, uint64_t diversifier);

/* 
 * ust_xfm_u16()
 *
 * implements the ust transform with 16-bit explicit seq num
 */

err_status_t
ust_xfm_u16(ust_ptr_t ctx,       /* pointer to ust context           */
	    uint16_t idx,        /* index                            */
	    octet_t *enc_start,  /* pointer to encryption start      */
	    int enc_len,         /* number of octets to encrypt      */
	    octet_t *auth_start, /* pointer to authentication start  */
	    int auth_len,        /* number of octets to authenticate */
	    octet_t *tag         /* authentication tag               */
	    );

/*
 * ust_inv_xfm_u16()
 * 
 * implements the ust inverse transform, with a 16-bit explicit index
 */

err_status_t
ust_inv_xfm_u16(ust_ptr_t ctx,       /* ust context                      */
		uint16_t idx,        /* explicit part of seq num         */
		octet_t *enc_start,  /* pointer to encryption start      */
		int enc_len,         /* number of octets to encrypt      */
		octet_t *auth_start, /* pointer to authentication start  */
		int auth_len,        /* number of octets to authenticate */
		octet_t *tag         /* authentication tag               */
		);

/* ust_ctx_t holds the context associated with a ust connection */

typedef struct ust_ctx_t {
  cipher_t *c;             /* cipher used as keystream generator     */
  auth_t   *h;             /* authentication function                */
  rdbx_t    rdbx;          /* replay database                        */
  octet_t  *auth_tag;      /* storage for authentication tags        */
} ust_ctx_t;


/*
 * ust_xfmv is an iovec - scatter/gather interface to ust
 * 
 * this interface is not yet supported in this implementation
 */

#if HAVE_SYS_UIO_H
#include <sys/uio.h>
#else

/* iovec isn't defined in the standard headers, so define it here */

struct iovec {
  octet_t *iov_base;
  size_t iov_len;
};

#endif 

err_status_t
ust_xfmv(ust_ptr_t ctx,          /* pointer to ust context                  */
         xtd_seq_num_t idx,      /* index                                   */
         struct iovec *enc_vec,  /* iovec of data to be encrypted           */
         int enc_vec_count,      /* length of iovec of data to be encrypted */
         struct iovec *auth_vec, /* iovec of data to be authenticated       */
         int auth_vec_count,     /* length of iovec of data to be authntctd */
         octet_t *tag            /* authentication tag                      */
         );

#endif /* UST_H */


