/*
 * ust_driver.c
 *
 * test driver for UST implementation
 *
 * David A. McGrew
 * Cisco Systems, Inc.
 */


#include <stdio.h>      /* for printf() */
#include <time.h>       /* for clock()  */
#include <unistd.h>     /* for getopt() */

#include "ust.h"

extern cipher_type_t null_cipher;
extern cipher_type_t rijndael_icm;
extern auth_type_t   null_auth;
extern auth_type_t   tmmhv2;

#define MAX_MSG_OCTETS 1024
#define NUM_TRIALS     100000

void
usage(char *prog_name) {
  printf("usage: %s [ -t ][ -v ][ -l ][ -d <debug_module> ]*\n"
	 "  -t         run timing test\n"
	 "  -v         run validation tests\n"
	 "  -d <mod>   turn on debugging module <mod>\n"
	 "  -l         list debugging modules\n", prog_name);

  exit(255);
}

double
ust_bits_per_second(ust_ctx_t *ctx, int msg_len_octets, int num_trials);

err_status_t
ust_test(ust_ctx_t *xfm_ctx, int msg_len_octets);

err_status_t
ust_auth_test();

int
main (int argc, char *argv[]) {
  octet_t cipher_key[30] = {
    0x43, 0x3e, 0x69, 0x7c, 0xf8, 0x9d, 0x46, 0x82,
    0xe8, 0x98, 0x76, 0x58, 0xd5, 0x65, 0xb9, 0xbc,
    0x43, 0x3e, 0x69, 0x7c, 0xf8, 0x9d, 0x46, 0x82,
    0xe8, 0x98, 0x76, 0x58, 0xd5, 0x65,
  };
 octet_t auth_key[94]   = {
   0x1f, 0x5a, 0x5f, 0xc0, 0x6b, 0x79, 0x4e, 0xbe,
   0x72, 0x38, 0x20, 0x53, 0x39, 0x5b, 0xbb, 0x9b,
   0xee, 0x22, 0x7c, 0x7f, 0x40, 0x3e, 0xb0, 0x5c,
   0xbd, 0x32, 0x3e, 0x80, 0x86, 0xa9, 0xad, 0x59,
   0x33, 0x19, 0x86, 0xcd, 0x25, 0x8b, 0xc6, 0x66,
   0xba, 0x45, 0x09, 0x6c, 0xdc, 0xaf, 0xa2, 0x49,
   0x81, 0x5f, 0xdb, 0x33, 0xa3, 0x43, 0xa7, 0x71,
   0xf8, 0xb6, 0xd1, 0xee, 0xb2, 0x23, 0xdf, 0xb4,
   0xb4, 0x0e, 0x5f, 0x10 
 };
  ust_ctx_t xfm_ctx;
  err_status_t stat;
  int len;
  char q;
  unsigned do_timing_test  = 0;
  unsigned do_validation   = 0;
  unsigned do_list_mods    = 0;

  if (argc == 1) 
    usage(argv[0]);

 /* initialize kernel - we need to do this before anything else */ 
  stat = crypto_kernel_init();
  if (stat) {
    printf("error: crypto_kernel init failed\n");
    exit(1);
  }
  printf("crypto_kernel successfully initalized\n");

  /* process input arguments */
  while (1) {
    q = getopt(argc, argv, "tvld:");
    if (q == -1) 
      break;
    switch (q) {
    case 't':
      do_timing_test = 1;
      break;
    case 'v':
      do_validation = 1;
      break;
    case 'l':
      do_list_mods = 1;
      break;
    case 'd':
      stat = crypto_kernel_set_debug_module(optarg, 1);
      if (stat) {
	printf("error: set debug module (%s) failed\n", optarg);
	exit(1);
      }
      break;
    default:
      usage(argv[0]);
    }    
  }

  if (!do_validation && !do_timing_test && ! do_list_mods)
    usage(argv[0]);

  if (do_list_mods) {
    stat = crypto_kernel_list_debug_modules();
    if (stat) {
      printf("error: list of debug modules failed\n");
      exit(1);
    }
  }    

  stat = ust_alloc(&xfm_ctx,       /* ust context                       */  
		   AES_128_ICM,    /* keystream generator identifier    */
		   30,             /* number of octets in cipher key    */
		   UST_TMMHv2,     /* auth algorithm identifier         */
		   94,             /* number of octets in the auth key  */
		   4,              /* number of octets in the auth tag  */
		   128             /* length of replay window           */
		   );
  if (stat) {
    fprintf(stderr, "ust_xfm failed with error code %d\n", stat);
    exit(1);
  }

  stat = ust_init(&xfm_ctx, cipher_key, auth_key);
  if (stat) {
    fprintf(stderr, "ust_xfm failed with error code %d\n", stat);
    exit(1);
  }

  if (do_validation) {
    printf("testing ust authentication function...");
    stat = ust_auth_test();
    if (stat) {
      fprintf(stderr, "ust_auth_test failed with error code %d\n", stat);
      exit(1);
    }
    printf("done\n");
    
    printf("testing ust_xfm and ust_inv_xfm...");
    stat = ust_test(&xfm_ctx, 64);  
    if (stat) {
      printf("ust_test failed with error code %d\n", stat);
      exit(1);
    } 
    printf("done\n");
  }

  if (do_timing_test) {
    printf("timing ust_xfm()...\n");
    for (len=16; len < MAX_MSG_OCTETS; len *= 2) 
      printf("%d octets\t%f bits per second\n", 
	     len, ust_bits_per_second(&xfm_ctx, len, NUM_TRIALS) / 1e9 );
  }

  return 0;
}

double
ust_bits_per_second(ust_ctx_t *xfm_ctx, int msg_len_octets, int num_trials) {
  clock_t timer;
  int i;
  xtd_seq_num_t idx;
  err_status_t stat;
  octet_t *msg; 


  /* allocate storage */
  msg = xalloc(msg_len_octets);
  if (msg == NULL)
    return 0.0;  /* signal failure by returning zero */
  
  /* re-initialize ust_ctx (to avoid replay protection) */
  rdbx_init(&xfm_ctx->rdbx);
  
  idx = 0;
  timer = clock();
  for(i=0; i < NUM_TRIALS; i++, index_advance(&idx, 1)) {
    stat = ust_xfm(xfm_ctx, idx, msg, msg_len_octets, 
		   msg, msg_len_octets, xfm_ctx->auth_tag);
    if (stat) {
      fprintf(stderr, "ust_xfm failed with error code %d\n", stat);
      exit(1);	
    }
  }
  timer = clock() - timer;

  free(msg);

  return (double) CLOCKS_PER_SEC * NUM_TRIALS * 8 * msg_len_octets / timer;
} 

/*
 * ust_test(ctx, len) creates a dummy message to which it applies 
 * a transform, then applies the inverse transform, verifying
 * that the data buffers before and after processing match, and that
 * the authentication works as well
 */

err_status_t
ust_test(ust_ctx_t *xfm_ctx, int msg_len_octets) {
  xtd_seq_num_t idx = 0;
  err_status_t stat;
  octet_t *msg; 
  octet_t *tag;
  int i;

  /* allocate storage, set message to all nulls */
  msg = xalloc(msg_len_octets + ust_get_tag_len(xfm_ctx));
  if (msg == NULL)
    return err_status_alloc_fail;
  /* the tag occupies the final portion of the newly allocated memory */
  tag = msg + msg_len_octets;
  for (i=0; i < msg_len_octets; i++)
    msg[i] = 0;

  /* check for proper decryption and false negatives */

  /* re-initialize ust_ctx (to avoid replay protection) */
  rdbx_init(&xfm_ctx->rdbx);
  idx = 0;

  /* apply transform to msg */
  stat = ust_xfm(xfm_ctx, idx, msg, msg_len_octets, 
		 msg, msg_len_octets, tag);
  if (stat) {
    free(msg);
    return stat;
  }

  /* re-initialize ust_ctx (to avoid replay protection) */
  rdbx_init(&xfm_ctx->rdbx);
  idx = 0;

  /* apply inverse transform to msg */
  stat = ust_inv_xfm(xfm_ctx, idx, msg, msg_len_octets, 
		     msg, msg_len_octets, tag);
  if (stat) {
    free(msg);
    return stat;
  }

  /* check that decrypted message is all-null */
  for (i=0; i < msg_len_octets; i++)
    if (msg[i] != 0x0000) {
      free(msg);
      return err_status_algo_fail;
    }

  /* now check for a false positive */

  /* re-initialize ust_ctx (to avoid replay protection) */
  rdbx_init(&xfm_ctx->rdbx);
  idx = 0;

  /* apply transform to msg */
  stat = ust_xfm(xfm_ctx, idx, NULL, msg_len_octets, 
		 msg, msg_len_octets, tag);
  if (stat) {
    free(msg);
    return stat;
  }

  /* re-initialize ust_ctx (to avoid replay protection) */
  rdbx_init(&xfm_ctx->rdbx);
  idx = 0;

  /* 
   * change message, then check for authentication failure
   */
  for (i=0; i < msg_len_octets; i++)
    msg[i] ^= 0xff;

  /* apply inverse transform to msg */
  stat = ust_inv_xfm(xfm_ctx, idx, NULL, msg_len_octets, 
		     msg, msg_len_octets, tag);
  if (stat != err_status_auth_fail) {
    free(msg);
    return err_status_algo_fail;
  }

  /* free memory and return success */
  free(msg);
  return err_status_ok;
} 


err_status_t
ust_auth_test() {
  octet_t auth_key[94]   = {
    0x23, 0x37, 0x47, 0xe0, 0x15, 0x64, 0x67, 0x1b,
    0x6f, 0x80, 0xdc, 0xdd, 0xa6, 0xcc, 0x5f, 0xf1,
    0x3e, 0x5d, 0x88, 0xeb, 0x61, 0x2e, 0x7c, 0x99,
    0x02, 0xe8, 0xd8, 0xb2, 0x77, 0xa5, 0x09, 0xf9,
    0xf0, 0xbc, 0x99, 0x97, 0x1d, 0xc9, 0xd4, 0x78,
    0x39, 0x6f, 0x96, 0x02, 0x85, 0x38, 0xaa, 0x7f,
    0x16, 0xa0, 0xa4, 0x56, 0xe7, 0x7e, 0x52, 0x62,
    0xa1, 0xdc, 0x6b, 0x06, 0x5e, 0x67, 0xb2, 0xd5,
    0x74, 0xee, 0x50, 0x45, 0x82, 0xc1, 0x31, 0x0a,
    0x28, 0xa5, 0xbb, 0x49, 0xf1, 0x5a, 0x38, 0x34,
    0x59, 0xc8, 0x0f, 0x3a, 0x36, 0xf7, 0x1b, 0x8c,
    0x95, 0x3d, 0xbf, 0x74, 0x20, 0x80 	   
  };
  octet_t message[56] = {
    0x57, 0x41, 0x52, 0x20, 0x49, 0x53, 0x20, 0x50,
    0x45, 0x41, 0x43, 0x45, 0x2c, 0x20, 0x46, 0x52,
    0x45, 0x45, 0x44, 0x4f, 0x4d, 0x20, 0x49, 0x53,
    0x20, 0x53, 0x4c, 0x41, 0x56, 0x45, 0x52, 0x59,
    0x2c, 0x20, 0x49, 0x47, 0x4e, 0x4f, 0x52, 0x41,
    0x4e, 0x43, 0x45, 0x20, 0x49, 0x53, 0x20, 0x53,
    0x54, 0x52, 0x45, 0x4e, 0x47, 0x54, 0x48, 0x00 
  };
  octet_t tag[4], ref_tag[4] = { 0x1D, 0x0F, 0x75, 0x36};
  xtd_seq_num_t idx = 0;
  ust_ctx_t xfm_ctx;
  err_status_t stat;

  stat = ust_alloc(&xfm_ctx,       /* ust context                       */  
		  NULL_CIPHER,     /* keystream generator identifier    */
		  0,               /* number of octets in cipher key    */
		  UST_TMMHv2,      /* auth algorithm identifier         */
		  94,              /* number of octets in the auth key  */
		  4,               /* number of octets in the auth tag  */
		  128              /* length of replay window           */
		  );
   
  if (stat)
    return err_status_algo_fail;

  stat = ust_init(&xfm_ctx, NULL, auth_key);
   
  if (stat)
    return err_status_algo_fail;
      
  stat = ust_xfm(&xfm_ctx, idx, NULL, 0, message, 56, tag); 
  if (stat) 
    return err_status_algo_fail; 

  if (octet_string_is_eq(tag, ref_tag, 4))
    return err_status_algo_fail;
  
  ust_inv_xfm(&xfm_ctx, idx, NULL, 0, message, 56, tag);
  if (stat)
    return err_status_algo_fail;
  
  return err_status_ok;
   
 }
